{% extends "!layout.html" %}
{%- block extrahead %}
{% if not embedded %}
<script src="https://ajax.googleapis.com/ajax/libs/jquery/3.3.1/jquery.min.js"></script>
<script type="text/javascript" src="{{ pathto('_static/copybutton.js', 1) }}"></script>
{% endif %}
{% endblock %}

{# astropy hardcodes a link to astropy.org in the header bar, which we don't want #}
{# It also hardcodes an Astropy logo in the `a.brand` element that we remove #}
{% block header %}
<div class="topbar">
  <a class="brand" title="{{ _('Documentation Home') }}" href="{{ pathto(master_doc) }}" style="background: transparent;"><span id="logotext1">{{ theme_logotext1 }}</span><span id="logotext2">{{ theme_logotext2 }}</span><span id="logotext3">{{ theme_logotext3 }}</span></a>
  <ul>
    <li><a title="{{ _('General Index') }}" href="{{ pathto('genindex') }}">Index</a></li>
    <li><a title="{{ _('Module Index') }}" href="{{ pathto('py-modindex') }}">Modules</a></li>
    <li>
      {% block sidebarsearch %}
      {% include "searchbox.html" %}
      {% endblock %}
    </li>
  </ul>
</div>
{% endblock %}
